@@ -1,21 +1,40 @@
-/*
-   +----------------------------------------------------------------------+
-   | PHP Version 5                                                        |
-   +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2002 The PHP Group                                |
-   +----------------------------------------------------------------------+
-   | This source file is subject to version 2.02 of the PHP license,      |
-   | that is bundled with this package in the file LICENSE, and is        |
-   | available at through the world-wide-web at                           |
-   | http://www.php.net/license/2_02.txt.                                 |
-   | If you did not receive a copy of the PHP license and are unable to   |
-   | obtain it through the world-wide-web, please send a note to          |
-   | license@php.net so we can mail you a copy immediately.               |
-   +----------------------------------------------------------------------+
-   | Authors: Matthew Sachs <matthewg@zevils.com>                         |
-   |          Eduardo Bacchi Kienetz <eduardo@kienetz.com>                |
-   |          Alfred Klomp <git@alfredklomp.com>                          |
-   +----------------------------------------------------------------------+
+/* ------------------------------------------------------------------
+ * This file is part of libsmbclient-php: Samba bindings for PHP.
+ * Libsmbclient-php is licensed under the BSD 2-clause license:
+ * ------------------------------------------------------------------
+ *
+ * Copyright (c) 2003,        Matthew Sachs
+ *               2009 - 2014, Eduardo Bacchi Kienetz
+ *               2013 - 2015, Alfred Klomp
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ *
+ * 2. Redistributions in binary form must reproduce the above
+ *    copyright notice, this list of conditions and the following
+ *    disclaimer in the documentation and/or other materials provided
+ *    with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
+ * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
+ * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
+ * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
+ * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
+ * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ *
+ * ------------------------------------------------------------------
  */
 
 #define IS_EXT_MODULE
@@ -30,7 +49,7 @@
 
 #include <libsmbclient.h>
 
-#define LIBSMBCLIENT_VERSION	"0.5.0"
+#define LIBSMBCLIENT_VERSION	"0.7.0"
 
 /* If Zend Thread Safety (ZTS) is defined, each thread gets its own copy of
  * the php_libsmbclient_globals structure. Else we use a single global copy: */
@@ -65,6 +84,28 @@ php_libsmbclient_state;
 static int le_libsmbclient_state;
 static int le_libsmbclient_file;
 
+enum {
+	SMBCLIENT_OPT_OPEN_SHAREMODE = 1,
+	SMBCLIENT_OPT_ENCRYPT_LEVEL = 2,
+	SMBCLIENT_OPT_CASE_SENSITIVE = 3,
+	SMBCLIENT_OPT_BROWSE_MAX_LMB_COUNT = 4,
+	SMBCLIENT_OPT_URLENCODE_READDIR_ENTRIES = 5,
+	/* Ignore OneSharePerServer, not relevant to us. */
+	SMBCLIENT_OPT_USE_KERBEROS = 6,
+	SMBCLIENT_OPT_FALLBACK_AFTER_KERBEROS = 7,
+	/* Reverse the sense of this option, the original
+	 * is the confusing "NoAutoAnonymousLogin": */
+	SMBCLIENT_OPT_AUTO_ANONYMOUS_LOGIN = 8,
+	SMBCLIENT_OPT_USE_CCACHE = 9,
+	SMBCLIENT_OPT_USE_NT_HASH = 10,
+	SMBCLIENT_OPT_NETBIOS_NAME = 11,
+	SMBCLIENT_OPT_WORKGROUP = 12,
+	SMBCLIENT_OPT_USER = 13,
+	SMBCLIENT_OPT_PORT = 14,
+	SMBCLIENT_OPT_TIMEOUT = 15,
+}
+php_libsmbclient_options;
+
 static char *
 find_char (char *start, char *last, char q)
 {
@@ -127,35 +168,149 @@ hide_password (char *url, int len)
 	astfill(second_colon + 1, at_sign - 1);
 }
 
+/* {{{ arginfo */
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_void, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_state_init, 0, 0, 1)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, workgroup)
+	ZEND_ARG_INFO(0, user)
+	ZEND_ARG_INFO(0, password)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_state, 0)
+	ZEND_ARG_INFO(0, state)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_option_get, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, option)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_option_set, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, option)
+	ZEND_ARG_INFO(0, value)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_path, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, path)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_rename, 0)
+	ZEND_ARG_INFO(0, oldstate)
+	ZEND_ARG_INFO(0, oldpath)
+	ZEND_ARG_INFO(0, oldstate)
+	ZEND_ARG_INFO(0, oldpath)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_dir, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, dir)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_file, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, file)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_open, 0, 0, 3)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, path)
+	ZEND_ARG_INFO(0, flags)
+	ZEND_ARG_INFO(0, mode)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_creat, 0, 0, 2)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, path)
+	ZEND_ARG_INFO(0, mode)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_write, 0, 0, 3)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, file)
+	ZEND_ARG_INFO(0, buffer)
+	ZEND_ARG_INFO(0, count)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_lseek, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, file)
+	ZEND_ARG_INFO(0, offset)
+	ZEND_ARG_INFO(0, whence)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_ftruncate, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, file)
+	ZEND_ARG_INFO(0, offset)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_utimes, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, path)
+	ZEND_ARG_INFO(0, mtime)
+	ZEND_ARG_INFO(0, atime)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO(arginfo_smbclient_getxattr, 0)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, path)
+	ZEND_ARG_INFO(0, name)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_setxattr, 0, 0, 4)
+	ZEND_ARG_INFO(0, state)
+	ZEND_ARG_INFO(0, path)
+	ZEND_ARG_INFO(0, name)
+	ZEND_ARG_INFO(0, value)
+	ZEND_ARG_INFO(0, flags)
+ZEND_END_ARG_INFO()
+
+/* }}} */
+
 static zend_function_entry libsmbclient_functions[] =
 {
-	PHP_FE(smbclient_state_new, NULL)
-	PHP_FE(smbclient_state_init, NULL)
-	PHP_FE(smbclient_state_errno, NULL)
-	PHP_FE(smbclient_state_free, NULL)
-	PHP_FE(smbclient_opendir, NULL)
-	PHP_FE(smbclient_readdir, NULL)
-	PHP_FE(smbclient_closedir, NULL)
-	PHP_FE(smbclient_stat, NULL)
-	PHP_FE(smbclient_fstat, NULL)
-	PHP_FE(smbclient_open, NULL)
-	PHP_FE(smbclient_creat, NULL)
-	PHP_FE(smbclient_read, NULL)
-	PHP_FE(smbclient_close, NULL)
-	PHP_FE(smbclient_mkdir, NULL)
-	PHP_FE(smbclient_rmdir, NULL)
-	PHP_FE(smbclient_rename, NULL)
-	PHP_FE(smbclient_write, NULL)
-	PHP_FE(smbclient_unlink, NULL)
-	PHP_FE(smbclient_lseek, NULL)
-	PHP_FE(smbclient_ftruncate, NULL)
-	PHP_FE(smbclient_chmod, NULL)
-	PHP_FE(smbclient_utimes, NULL)
-	PHP_FE(smbclient_listxattr, NULL)
-	PHP_FE(smbclient_getxattr, NULL)
-	PHP_FE(smbclient_setxattr, NULL)
-	PHP_FE(smbclient_removexattr, NULL)
+	PHP_FE(smbclient_version, arginfo_smbclient_void)
+	PHP_FE(smbclient_library_version, arginfo_smbclient_void)
+	PHP_FE(smbclient_state_new, arginfo_smbclient_void)
+	PHP_FE(smbclient_state_init, arginfo_smbclient_state_init)
+	PHP_FE(smbclient_state_errno, arginfo_smbclient_state)
+	PHP_FE(smbclient_state_free, arginfo_smbclient_state)
+	PHP_FE(smbclient_option_get, arginfo_smbclient_option_get)
+	PHP_FE(smbclient_option_set, arginfo_smbclient_option_set)
+	PHP_FE(smbclient_opendir, arginfo_smbclient_path)
+	PHP_FE(smbclient_readdir, arginfo_smbclient_dir)
+	PHP_FE(smbclient_closedir, arginfo_smbclient_dir)
+	PHP_FE(smbclient_stat, arginfo_smbclient_path)
+	PHP_FE(smbclient_fstat, arginfo_smbclient_file)
+	PHP_FE(smbclient_open, arginfo_smbclient_open)
+	PHP_FE(smbclient_creat, arginfo_smbclient_creat)
+	PHP_FE(smbclient_read, arginfo_smbclient_file)
+	PHP_FE(smbclient_close, arginfo_smbclient_file)
+	PHP_FE(smbclient_mkdir, arginfo_smbclient_creat)
+	PHP_FE(smbclient_rmdir, arginfo_smbclient_path)
+	PHP_FE(smbclient_rename, arginfo_smbclient_rename)
+	PHP_FE(smbclient_write, arginfo_smbclient_write)
+	PHP_FE(smbclient_unlink, arginfo_smbclient_path)
+	PHP_FE(smbclient_lseek, arginfo_smbclient_lseek)
+	PHP_FE(smbclient_ftruncate, arginfo_smbclient_ftruncate)
+	PHP_FE(smbclient_chmod, arginfo_smbclient_creat)
+	PHP_FE(smbclient_utimes, arginfo_smbclient_utimes)
+	PHP_FE(smbclient_listxattr, arginfo_smbclient_path)
+	PHP_FE(smbclient_getxattr, arginfo_smbclient_getxattr)
+	PHP_FE(smbclient_setxattr, arginfo_smbclient_setxattr)
+	PHP_FE(smbclient_removexattr, arginfo_smbclient_getxattr)
+	PHP_FE(smbclient_statvfs, arginfo_smbclient_path)
+	PHP_FE(smbclient_fstatvfs, arginfo_smbclient_file)
+#ifdef PHP_FE_END
+	PHP_FE_END
+#else
 	{NULL, NULL, NULL}
+#endif
 };
 
 zend_module_entry libsmbclient_module_entry =
@@ -265,6 +420,44 @@ PHP_MINIT_FUNCTION(smbclient)
 	REGISTER_LONG_CONSTANT("SMBCLIENT_XATTR_CREATE", SMBC_XATTR_FLAG_CREATE, CONST_PERSISTENT | CONST_CS);
 	REGISTER_LONG_CONSTANT("SMBCLIENT_XATTR_REPLACE", SMBC_XATTR_FLAG_REPLACE, CONST_PERSISTENT | CONST_CS);
 
+	/* Constants for getting/setting options: */
+	#define SMBCLIENT_CONST(x) REGISTER_LONG_CONSTANT(#x, x, CONST_PERSISTENT | CONST_CS);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_OPEN_SHAREMODE);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_ENCRYPT_LEVEL);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_CASE_SENSITIVE);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_BROWSE_MAX_LMB_COUNT);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_URLENCODE_READDIR_ENTRIES);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_USE_KERBEROS);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_FALLBACK_AFTER_KERBEROS);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_AUTO_ANONYMOUS_LOGIN);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_USE_CCACHE);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_USE_NT_HASH);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_NETBIOS_NAME);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_WORKGROUP);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_USER);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_PORT);
+	SMBCLIENT_CONST(SMBCLIENT_OPT_TIMEOUT);
+	#undef SMBCLIENT_CONST
+
+	/* Constants for use with SMBCLIENT_OPT_OPENSHAREMODE: */
+	REGISTER_LONG_CONSTANT("SMBCLIENT_SHAREMODE_DENY_DOS", SMBC_SHAREMODE_DENY_DOS, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_SHAREMODE_DENY_ALL", SMBC_SHAREMODE_DENY_ALL, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_SHAREMODE_DENY_WRITE", SMBC_SHAREMODE_DENY_WRITE, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_SHAREMODE_DENY_READ", SMBC_SHAREMODE_DENY_READ, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_SHAREMODE_DENY_NONE", SMBC_SHAREMODE_DENY_NONE, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_SHAREMODE_DENY_FCB", SMBC_SHAREMODE_DENY_FCB, CONST_PERSISTENT | CONST_CS);
+
+	/* Constants for use with SMBCLIENT_OPT_ENCRYPTLEVEL: */
+	REGISTER_LONG_CONSTANT("SMBCLIENT_ENCRYPTLEVEL_NONE", SMBC_ENCRYPTLEVEL_NONE, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_ENCRYPTLEVEL_REQUEST", SMBC_ENCRYPTLEVEL_REQUEST, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_ENCRYPTLEVEL_REQUIRE", SMBC_ENCRYPTLEVEL_REQUIRE, CONST_PERSISTENT | CONST_CS);
+
+	/* Constants for the VFS functions: */
+	REGISTER_LONG_CONSTANT("SMBCLIENT_VFS_RDONLY", SMBC_VFS_FEATURE_RDONLY, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_VFS_DFS", SMBC_VFS_FEATURE_DFS, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_VFS_CASE_INSENSITIVE", SMBC_VFS_FEATURE_CASE_INSENSITIVE, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("SMBCLIENT_VFS_NO_UNIXCIFS", SMBC_VFS_FEATURE_NO_UNIXCIFS, CONST_PERSISTENT | CONST_CS);
+
 	le_libsmbclient_state = zend_register_list_destructors_ex(smbclient_state_dtor, NULL, PHP_LIBSMBCLIENT_STATE_NAME, module_number);
 	le_libsmbclient_file = zend_register_list_destructors_ex(smbclient_file_dtor, NULL, PHP_LIBSMBCLIENT_FILE_NAME, module_number);
 
@@ -285,7 +478,8 @@ PHP_MINFO_FUNCTION(smbclient)
 {
 	php_info_print_table_start();
 	php_info_print_table_row(2, "libsmbclient Support", "enabled");
-	php_info_print_table_row(2, "libsmbclient Version", LIBSMBCLIENT_VERSION);
+	php_info_print_table_row(2, "libsmbclient extension Version", LIBSMBCLIENT_VERSION);
+	php_info_print_table_row(2, "libsmbclient library Version", smbc_version());
 	php_info_print_table_end();
 }
 
@@ -294,6 +488,9 @@ PHP_FUNCTION(smbclient_state_new)
 	SMBCCTX *ctx;
 	php_libsmbclient_state *state;
 
+	if (zend_parse_parameters_none() == FAILURE) {
+		RETURN_FALSE;
+	}
 	if ((ctx = smbc_new_context()) == NULL) {
 		switch (errno) {
 			case ENOMEM: php_error(E_WARNING, "Couldn't create smbclient state: insufficient memory"); break;
@@ -317,6 +514,9 @@ PHP_FUNCTION(smbclient_state_new)
 	 * find the state from the context in the auth function: */
 	smbc_setOptionUserData(ctx, (void *)state);
 
+	/* Force full, modern timenames when getting xattrs: */
+	smbc_setOptionFullTimeNames(state->ctx, 1);
+
 	ZEND_REGISTER_RESOURCE(return_value, state, le_libsmbclient_state);
 }
 
@@ -355,6 +555,22 @@ ctx_init_getauth (zval *z, char **dest, int *destlen, char *varname)
 	}
 }
 
+PHP_FUNCTION(smbclient_version)
+{
+	if (zend_parse_parameters_none() == FAILURE) {
+		RETURN_FALSE;
+	}
+	RETURN_STRING(LIBSMBCLIENT_VERSION, 1);
+}
+
+PHP_FUNCTION(smbclient_library_version)
+{
+	if (zend_parse_parameters_none() == FAILURE) {
+		RETURN_FALSE;
+	}
+	RETURN_STRING(smbc_version(), 1);
+}
+
 PHP_FUNCTION(smbclient_state_init)
 {
 	SMBCCTX *ctx;
@@ -483,10 +699,26 @@ PHP_FUNCTION(smbclient_opendir)
 	RETURN_FALSE;
 }
 
+static char *
+type_to_string (unsigned int type)
+{
+	switch (type) {
+		case SMBC_WORKGROUP	: return "workgroup";
+		case SMBC_SERVER	: return "server";
+		case SMBC_FILE_SHARE	: return "file share";
+		case SMBC_PRINTER_SHARE	: return "printer share";
+		case SMBC_COMMS_SHARE	: return "communication share";
+		case SMBC_IPC_SHARE	: return "IPC share";
+		case SMBC_DIR		: return "directory";
+		case SMBC_FILE		: return "file";
+		case SMBC_LINK		: return "link";
+	}
+	return "unknown";
+}
+
 PHP_FUNCTION(smbclient_readdir)
 {
 	struct smbc_dirent *dirent;
-	char *type;
 	zval *zstate;
 	zval *zfile;
 	SMBCFILE *file;
@@ -516,19 +748,7 @@ PHP_FUNCTION(smbclient_readdir)
 		php_error(E_WARNING, "Couldn't initialize array");
 		RETURN_FALSE;
 	}
-	switch (dirent->smbc_type) {
-		case SMBC_WORKGROUP: type = "workgroup"; break;
-		case SMBC_SERVER: type = "server"; break;
-		case SMBC_FILE_SHARE: type = "file share"; break;
-		case SMBC_PRINTER_SHARE: type = "printer share"; break;
-		case SMBC_COMMS_SHARE: type = "communication share"; break;
-		case SMBC_IPC_SHARE: type = "IPC share"; break;
-		case SMBC_DIR: type = "directory"; break;
-		case SMBC_FILE: type = "file"; break;
-		case SMBC_LINK: type = "link"; break;
-		default: type = "unknown"; break;
-	}
-	add_assoc_string(return_value, "type", type, 1);
+	add_assoc_string(return_value, "type", type_to_string(dirent->smbc_type), 1);
 	add_assoc_stringl(return_value, "comment", dirent->comment, dirent->commentlen, 1);
 	add_assoc_stringl(return_value, "name", dirent->name, dirent->namelen, 1);
 }
@@ -1341,3 +1561,285 @@ PHP_FUNCTION(smbclient_removexattr)
 	}
 	RETURN_FALSE;
 }
+
+PHP_FUNCTION(smbclient_option_get)
+{
+	long option;
+	char *ret;
+	zval *zstate;
+	php_libsmbclient_state *state;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &zstate, &option) == FAILURE) {
+		return;
+	}
+	STATE_FROM_ZSTATE;
+
+	switch (option)
+	{
+	case SMBCLIENT_OPT_OPEN_SHAREMODE:
+		RETURN_LONG(smbc_getOptionOpenShareMode(state->ctx));
+
+	case SMBCLIENT_OPT_ENCRYPT_LEVEL:
+		RETURN_LONG(smbc_getOptionSmbEncryptionLevel(state->ctx));
+
+	case SMBCLIENT_OPT_CASE_SENSITIVE:
+		RETURN_BOOL(smbc_getOptionCaseSensitive(state->ctx));
+
+	case SMBCLIENT_OPT_BROWSE_MAX_LMB_COUNT:
+		RETURN_LONG(smbc_getOptionBrowseMaxLmbCount(state->ctx));
+
+	case SMBCLIENT_OPT_URLENCODE_READDIR_ENTRIES:
+		RETURN_BOOL(smbc_getOptionUrlEncodeReaddirEntries(state->ctx));
+
+	case SMBCLIENT_OPT_USE_KERBEROS:
+		RETURN_BOOL(smbc_getOptionUseKerberos(state->ctx));
+
+	case SMBCLIENT_OPT_FALLBACK_AFTER_KERBEROS:
+		RETURN_BOOL(smbc_getOptionFallbackAfterKerberos(state->ctx));
+
+	/* Reverse the sense of this option, the original is confusing: */
+	case SMBCLIENT_OPT_AUTO_ANONYMOUS_LOGIN:
+		RETURN_BOOL(!(smbc_getOptionNoAutoAnonymousLogin(state->ctx)));
+
+	case SMBCLIENT_OPT_USE_CCACHE:
+		RETURN_BOOL(smbc_getOptionUseCCache(state->ctx));
+
+	#ifdef HAVE_SMBC_SETOPTIONUSENTHASH
+	case SMBCLIENT_OPT_USE_NT_HASH:
+		RETURN_BOOL(smbc_getOptionUseNTHash(state->ctx));
+	#endif
+
+	#ifdef HAVE_SMBC_SETPORT
+	case SMBCLIENT_OPT_PORT:
+		RETURN_LONG(smbc_getPort(state->ctx));
+	#endif
+
+	case SMBCLIENT_OPT_TIMEOUT:
+		RETURN_LONG(smbc_getTimeout(state->ctx));
+
+	case SMBCLIENT_OPT_NETBIOS_NAME:
+		if ((ret = smbc_getNetbiosName(state->ctx)) == NULL) {
+			RETURN_EMPTY_STRING();
+		}
+		if (strlen(ret) == 0) {
+			RETURN_EMPTY_STRING();
+		}
+		RETURN_STRING(ret, 1);
+
+	case SMBCLIENT_OPT_WORKGROUP:
+		if ((ret = smbc_getWorkgroup(state->ctx)) == NULL) {
+			RETURN_EMPTY_STRING();
+		}
+		if (strlen(ret) == 0) {
+			RETURN_EMPTY_STRING();
+		}
+		RETURN_STRING(ret, 1);
+
+	case SMBCLIENT_OPT_USER:
+		if ((ret = smbc_getUser(state->ctx)) == NULL) {
+			RETURN_EMPTY_STRING();
+		}
+		if (strlen(ret) == 0) {
+			RETURN_EMPTY_STRING();
+		}
+		RETURN_STRING(ret, 1);
+	}
+	RETURN_NULL();
+}
+
+PHP_FUNCTION(smbclient_option_set)
+{
+	long option;
+	zval *zstate;
+	zval *zvalue;
+	php_libsmbclient_state *state;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rlz", &zstate, &option, &zvalue) == FAILURE) {
+		return;
+	}
+	STATE_FROM_ZSTATE;
+
+	switch (Z_TYPE_P(zvalue))
+	{
+	case IS_BOOL:
+
+		switch (option)
+		{
+		case SMBCLIENT_OPT_CASE_SENSITIVE:
+			smbc_setOptionCaseSensitive(state->ctx, Z_BVAL_P(zvalue));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_URLENCODE_READDIR_ENTRIES:
+			smbc_setOptionUrlEncodeReaddirEntries(state->ctx, Z_BVAL_P(zvalue));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_USE_KERBEROS:
+			smbc_setOptionUseKerberos(state->ctx, Z_BVAL_P(zvalue));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_FALLBACK_AFTER_KERBEROS:
+			smbc_setOptionFallbackAfterKerberos(state->ctx, Z_BVAL_P(zvalue));
+			RETURN_TRUE;
+
+		/* Reverse the sense of this option: */
+		case SMBCLIENT_OPT_AUTO_ANONYMOUS_LOGIN:
+			smbc_setOptionNoAutoAnonymousLogin(state->ctx, !(Z_BVAL_P(zvalue)));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_USE_CCACHE:
+			smbc_setOptionUseCCache(state->ctx, Z_BVAL_P(zvalue));
+			RETURN_TRUE;
+
+		#ifdef HAVE_SMBC_SETOPTIONUSENTHASH
+		case SMBCLIENT_OPT_USE_NT_HASH:
+			smbc_setOptionUseNTHash(state->ctx, Z_BVAL_P(zvalue));
+			RETURN_TRUE;
+		#endif
+		}
+		break;
+
+	case IS_LONG:
+
+		switch (option)
+		{
+		case SMBCLIENT_OPT_OPEN_SHAREMODE:
+			smbc_setOptionOpenShareMode(state->ctx, Z_LVAL_P(zvalue));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_ENCRYPT_LEVEL:
+			smbc_setOptionSmbEncryptionLevel(state->ctx, Z_LVAL_P(zvalue));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_BROWSE_MAX_LMB_COUNT:
+			smbc_setOptionBrowseMaxLmbCount(state->ctx, Z_LVAL_P(zvalue));
+			RETURN_TRUE;
+
+		#ifdef HAVE_SMBC_SETPORT
+		case SMBCLIENT_OPT_PORT:
+			smbc_setPort(state->ctx, Z_LVAL_P(zvalue));
+			RETURN_TRUE;
+		#endif
+
+		case SMBCLIENT_OPT_TIMEOUT:
+			smbc_setTimeout(state->ctx, Z_LVAL_P(zvalue));
+			RETURN_TRUE;
+		}
+		break;
+
+	case IS_STRING:
+
+		switch (option)
+		{
+		case SMBCLIENT_OPT_NETBIOS_NAME:
+			smbc_setNetbiosName(state->ctx, Z_STRVAL_P(zvalue));
+			RETURN_TRUE;
+
+		/* For the next two options, update our state object as well: */
+		case SMBCLIENT_OPT_WORKGROUP:
+			if (ctx_init_getauth(zvalue, &state->wrkg, &state->wrkglen, "workgroup") == 0) {
+				RETURN_FALSE;
+			}
+			smbc_setWorkgroup(state->ctx, Z_STRVAL_P(zvalue));
+			RETURN_TRUE;
+
+		case SMBCLIENT_OPT_USER:
+			if (ctx_init_getauth(zvalue, &state->user, &state->userlen, "username") == 0) {
+				RETURN_FALSE;
+			}
+			smbc_setUser(state->ctx, Z_STRVAL_P(zvalue));
+			RETURN_TRUE;
+		}
+		break;
+	}
+	RETURN_FALSE;
+}
+
+PHP_FUNCTION(smbclient_statvfs)
+{
+	char *url;
+	int url_len;
+	zval *zstate;
+	struct statvfs st;
+	smbc_statvfs_fn smbc_statvfs;
+	php_libsmbclient_state *state;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zstate, &url, &url_len) == FAILURE) {
+		return;
+	}
+	STATE_FROM_ZSTATE;
+
+	if ((smbc_statvfs = smbc_getFunctionStatVFS(state->ctx)) == NULL) {
+		RETURN_FALSE;
+	}
+	if (smbc_statvfs(state->ctx, url, &st) != 0) {
+		hide_password(url, url_len);
+		switch (state->err = errno) {
+			case EBADF: php_error(E_WARNING, "Couldn't statvfs %s: bad file descriptor", url); break;
+			case EACCES: php_error(E_WARNING, "Couldn't statvfs %s: permission denied", url); break;
+			case EINVAL: php_error(E_WARNING, "Couldn't statvfs %s: library not initalized or otherwise invalid", url); break;
+			case ENOMEM: php_error(E_WARNING, "Couldn't statvfs %s: out of memory", url); break;
+			default: php_error(E_WARNING, "Couldn't statvfs %s: unknown error (%d)", url, errno); break;
+		}
+		RETURN_FALSE;
+	}
+	if (array_init(return_value) != SUCCESS) {
+		php_error(E_WARNING, "Couldn't initialize array");
+		RETURN_FALSE;
+	}
+	add_assoc_long(return_value, "bsize",   st.f_bsize);
+	add_assoc_long(return_value, "frsize",  st.f_frsize);
+	add_assoc_long(return_value, "blocks",  st.f_blocks);
+	add_assoc_long(return_value, "bfree",   st.f_bfree);
+	add_assoc_long(return_value, "bavail",  st.f_bavail);
+	add_assoc_long(return_value, "files",   st.f_files);
+	add_assoc_long(return_value, "ffree",   st.f_ffree);
+	add_assoc_long(return_value, "favail",  st.f_favail);
+	add_assoc_long(return_value, "fsid",    st.f_fsid);
+	add_assoc_long(return_value, "flag",    st.f_flag);
+	add_assoc_long(return_value, "namemax", st.f_namemax);
+}
+
+PHP_FUNCTION(smbclient_fstatvfs)
+{
+	zval *zstate;
+	zval *zfile;
+	SMBCFILE *file;
+	struct statvfs st;
+	smbc_fstatvfs_fn smbc_fstatvfs;
+	php_libsmbclient_state *state;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr", &zstate, &zfile) == FAILURE) {
+		return;
+	}
+	STATE_FROM_ZSTATE;
+	FILE_FROM_ZFILE;
+
+	if ((smbc_fstatvfs = smbc_getFunctionFstatVFS(state->ctx)) == NULL) {
+		RETURN_FALSE;
+	}
+	if (smbc_fstatvfs(state->ctx, file, &st) != 0) {
+		switch (state->err = errno) {
+			case EBADF: php_error(E_WARNING, "Couldn't fstatvfs: bad file descriptor"); break;
+			case EACCES: php_error(E_WARNING, "Couldn't fstatvfs: permission denied"); break;
+			case EINVAL: php_error(E_WARNING, "Couldn't fstatvfs: library not initalized or otherwise invalid"); break;
+			case ENOMEM: php_error(E_WARNING, "Couldn't fstatvfs: out of memory"); break;
+			default: php_error(E_WARNING, "Couldn't fstatvfs: unknown error (%d)", errno); break;
+		}
+		RETURN_FALSE;
+	}
+	if (array_init(return_value) != SUCCESS) {
+		php_error(E_WARNING, "Couldn't initialize array");
+		RETURN_FALSE;
+	}
+	add_assoc_long(return_value, "bsize",   st.f_bsize);
+	add_assoc_long(return_value, "frsize",  st.f_frsize);
+	add_assoc_long(return_value, "blocks",  st.f_blocks);
+	add_assoc_long(return_value, "bfree",   st.f_bfree);
+	add_assoc_long(return_value, "bavail",  st.f_bavail);
+	add_assoc_long(return_value, "files",   st.f_files);
+	add_assoc_long(return_value, "ffree",   st.f_ffree);
+	add_assoc_long(return_value, "favail",  st.f_favail);
+	add_assoc_long(return_value, "fsid",    st.f_fsid);
+	add_assoc_long(return_value, "flag",    st.f_flag);
+	add_assoc_long(return_value, "namemax", st.f_namemax);
+}
